@@ -4,6 +4,7 @@ import torch
 from sentence_transformers import SentenceTransformer
 from sklearn.cluster import KMeans
 from sklearn.decomposition import PCA, TruncatedSVD
+from sklearn.metrics import normalized_mutual_info_score
 from sklearn.preprocessing import MinMaxScaler
 from torch.utils.data import DataLoader
 from typing import List
@@ -271,26 +272,31 @@ class STC(torch.nn.Module):
     Class for building self-taught short text clustering using auto-encoder and k-means
     """
     def __init__(self,
-                 encoder_hidden_layer: dict,
-                 decoder_hidden_layer: dict,
                  dimensions: List[int],
                  iterator: DataLoader,
-                 batch_size: int,
+                 encoder_hidden_layer: dict = None,
+                 decoder_hidden_layer: dict = None,
+                 batch_size: int = 16,
                  epoch: int = 15,
-                 optimizer: str = 'rmsprop',
+                 optimizer: str = 'sgd',
                  n_clusters: int = 20,
                  alpha: float = 1.0,
                  n_iterations: int = 100,
-                 max_iterations: int = 15,
-                 tol: float = 0.001,
-                 update_interval: int = 140
+                 max_iterations: int = 1000,
+                 early_stopping: bool = False,
+                 tol: float = 0.0001,
+                 update_interval: int = 500
                  ):
         """
         :param encoder_hidden_layer: dict
-            Input and output dimensions for each encoder hidden layer
+            Input and output dimensions for each encoder hidden layer (input & output layer included)
+                -> e.g. for each layer: dict(layer_1=in_features=x, out_features=y)
+                -> layer_1 is always the input layer & the last layer is always the output layer
 
         :param decoder_hidden_layer: dict
-            Input and output dimensions for each decoder hidden layer
+            Input and output dimensions for each decoder hidden layer (input & output layer included)
+                -> e.g. for each layer: dict(layer_1=in_features=x, out_features=y)
+                -> layer_1 is always the input layer & the last layer is always the output layer
 
         :param dimensions: List[int]
             Number of dimensions of the encoder layers
@@ -322,6 +328,9 @@ class STC(torch.nn.Module):
         :param max_iterations: int
             Maximum number of iterations for running training
 
+        :param early_stopping: bool
+            Whether to stop early if stopping criterion matches or not
+
         :param tol: float
             Tolerance for
 
@@ -335,29 +344,50 @@ class STC(torch.nn.Module):
         self.n_iterations: int = n_iterations
         self.max_iterations: int = max_iterations
         self.update_interval: int = update_interval
+        self.early_stopping: bool = early_stopping
         self.tol: float = tol
         self.epoch: int = epoch
         self.batch_size: int = batch_size
         self.optimizer: str = optimizer
         self.iterator: DataLoader = iterator
+        self.nmi_score: float = 0.0
+        self.obs: List[int] = []
+        self.pred: List[str] = []
+        if encoder_hidden_layer is None or decoder_hidden_layer is None:
+            _encoder_hidden_layer: dict = dict(layer_1=dict(in_features=self.dims[1], out_features=500),
+                                               layer_2=dict(in_features=500, out_features=500),
+                                               layer_3=dict(in_features=500, out_features=2000),
+                                               layer_4=dict(in_features=2000, out_features=self.n_clusters),
+                                               )
+            _decoder_hidden_layer: dict = dict(layer_1=dict(in_features=self.n_clusters, out_features=2000),
+                                               layer_2=dict(in_features=2000, out_features=500),
+                                               layer_3=dict(in_features=500, out_features=500),
+                                               layer_4=dict(in_features=500, out_features=self.dims[1])
+                                               )
+        else:
+            _encoder_hidden_layer: dict = encoder_hidden_layer
+            _decoder_hidden_layer: dict = decoder_hidden_layer
         self.model: torch.nn.Sequential = torch.nn.Sequential()
-        self.encoder = Encoder(encoder_hidden_layer=encoder_hidden_layer)
-        self.decoder = Decoder(decoder_hidden_layer=decoder_hidden_layer)
-        self.auto_encoder = AutoEncoder(encoder_hidden_layer=encoder_hidden_layer, decoder_hidden_layer=decoder_hidden_layer)
-        print('Pre-train AutoEncoder ...')
+        self.encoder = Encoder(encoder_hidden_layer=_encoder_hidden_layer)
+        self.decoder = Decoder(decoder_hidden_layer=_decoder_hidden_layer)
+        self.auto_encoder = AutoEncoder(encoder_hidden_layer=_encoder_hidden_layer, decoder_hidden_layer=_decoder_hidden_layer)
         self.pre_train_auto_encoder()
         _clustering_layer = Clustering(dims=self.dims[1], n_clusters=self.n_clusters, alpha=self.alpha)
         self.model.add_module(name='encoder', module=self.encoder)
         self.model.add_module(name='cluster', module=_clustering_layer)
 
-    def _clip_gradient(self, clip_value: float):
+    def _clip_gradient(self, clip_value: float, auto_encoder: bool):
         """
         Clip gradient during network training
 
         :param clip_value: float
             Clipping value for avoiding vanishing gradient
+
+        :param auto_encoder: bool
+            Whether to use auto-encoder model parameters or not
         """
-        params = list(filter(lambda p: p.grad is not None, self.auto_encoder.parameters()))
+        _parameters = self.auto_encoder.parameters() if auto_encoder else self.model.parameters()
+        params = list(filter(lambda p: p.grad is not None, _parameters))
         for p in params:
             p.grad.data.clamp_(-clip_value, clip_value)
 
@@ -373,66 +403,63 @@ class STC(torch.nn.Module):
         """
         return self.encoder(x)
 
-    def fit(self, x: torch.tensor) -> tuple:
+    def fit(self, x: torch.tensor):
         """
         Train STC using PyTorch and KMeans from sklearn
 
         :param x: torch.tensor
             Embedding values
-
-        :return: tuple
-            Observed label and predicted label
         """
-        _predictions: List[int] = []
-        _observations: List[int] = []
         if torch.cuda.is_available():
             self.model.cuda()
         self.model.train()
         with torch.autograd.set_detect_anomaly(True):
             if self.optimizer == 'rmsprop':
-                _optim: torch.optim = torch.optim.RMSprop(params=filter(lambda p: p.requires_grad, self.model.parameters()), lr=0.0001)
+                _optim: torch.optim = torch.optim.RMSprop(params=filter(lambda p: p.requires_grad, self.model.parameters()), lr=0.01)
             elif self.optimizer == 'adam':
-                _optim: torch.optim = torch.optim.Adam(params=self.model.parameters(), lr=0.0001)
+                _optim: torch.optim = torch.optim.Adam(params=filter(lambda p: p.requires_grad, self.model.parameters()), lr=0.01)
             elif self.optimizer == 'sgd':
-                _optim: torch.optim = torch.optim.SGD(params=self.model.parameters(), lr=0.0001)
+                _optim: torch.optim = torch.optim.SGD(params=filter(lambda p: p.requires_grad, self.model.parameters()), lr=0.01)
             else:
                 raise SelfTaughtShortTextClusteringException('Optimizer ({}) not supported'.format(self.optimizer))
-            _loss_function: torch.nn.KLDivLoss = torch.nn.KLDivLoss()
+            _loss_function: torch.nn.KLDivLoss = torch.nn.KLDivLoss(reduction='batchmean')
             _k_means: KMeans = KMeans(n_clusters=self.n_clusters, n_init=self.n_iterations)
             _pred: np.ndarray = _k_means.fit_predict(self.encoder(x).detach().numpy())
             _last_pred: np.ndarray = np.copy(_pred)
+            _last_pred_tensor: torch.tensor = torch.tensor(data=_last_pred)
             self.model[1].set_weights(cluster_centers=_k_means.cluster_centers_)
             _index: int = 0
             _index_array: np.ndarray = np.arange(x.shape[0])
             for i in range(0, self.max_iterations, 1):
-                print('Iteration: ', i)
+                _idx: np.ndarray = _index_array[_index * self.batch_size: min((_index + 1) * self.batch_size, x.shape[0])]
+                _predictors, _target = self.iterator.dataset[_idx]
                 if i % self.update_interval == 0:
                     _q = self.model(x)
                     _p = self.target_distribution(_q)
                     _pred = _q.argmax(1)
-                    # check stopping criterion
                     _delta_label = np.sum(_pred != _last_pred).astype(np.float32) / _pred.shape[0]
                     _last_pred = np.copy(_pred)
-                    if i > 0 and _delta_label < self.tol:
-                        break
-                _idx: np.ndarray = _index_array[_index * self.batch_size: min((_index + 1) * self.batch_size, x.shape[0])]
-                _predictors, _target = self.iterator.dataset[_idx]
-                if torch.cuda.is_available():
-                    _predictors = _predictors.cuda()
-                _prediction = torch.softmax(input=self.model(_predictors), dim=0)
-                print('Predictions', _prediction)
-                _top_class, _top_p = _prediction.topk(k=1, dim=0)
-                print('Top Class', _top_class)
-                print('Top Prob', _top_p)
-                _predictions.extend(_top_class.detach().tolist())
-                _observations.extend(_target.detach().numpy().tolist())
-                _loss = _loss_function(_prediction, _p[_idx].detach())
+                    _last_pred_tensor = torch.tensor(data=_last_pred)
+                    for idx in enumerate(_idx):
+                        self.obs.append(_target.detach().numpy().tolist()[idx[0]])
+                        self.pred.append(_last_pred_tensor.detach().numpy().tolist()[idx[0]])
+                    self.fitness()
+                    if self.early_stopping:
+                        if i > 0 and _delta_label < self.tol:
+                            break
+                _loss = _loss_function(_q[_idx].detach(), _p[_idx].detach())
+                _loss.requires_grad = True
                 _optim.zero_grad()
                 _loss.backward(retain_graph=True)
-                self._clip_gradient(clip_value=1e-1)
+                self._clip_gradient(clip_value=1e-1, auto_encoder=False)
                 _optim.step()
                 _index = _index + 1 if (_index + 1) * self.batch_size <= x.shape[0] else 0
-        return _observations, _predictions
+
+    def fitness(self):
+        """
+        Fitness score (Normalized Mutual Information)
+        """
+        self.nmi_score: float = normalized_mutual_info_score(labels_true=self.obs, labels_pred=self.pred)
 
     def predict(self, x: np.ndarray) -> int:
         """
@@ -455,16 +482,15 @@ class STC(torch.nn.Module):
             self.auto_encoder.cuda()
         self.auto_encoder.train()
         if self.optimizer == 'rmsprop':
-            _optim: torch.optim = torch.optim.RMSprop(params=filter(lambda p: p.requires_grad, self.auto_encoder.parameters()), lr=0.0001)
+            _optim: torch.optim = torch.optim.RMSprop(params=filter(lambda p: p.requires_grad, self.auto_encoder.parameters()), lr=0.01)
         elif self.optimizer == 'adam':
-            _optim: torch.optim = torch.optim.Adam(params=self.auto_encoder.parameters(), lr=0.0001)
+            _optim: torch.optim = torch.optim.Adam(params=filter(lambda p: p.requires_grad, self.auto_encoder.parameters()), lr=0.01)
         elif self.optimizer == 'sgd':
-            _optim: torch.optim = torch.optim.SGD(params=self.auto_encoder.parameters(), lr=0.0001)
+            _optim: torch.optim = torch.optim.SGD(params=filter(lambda p: p.requires_grad, self.auto_encoder.parameters()), lr=0.01)
         else:
             raise SelfTaughtShortTextClusteringException('Optimizer ({}) not supported'.format(self.optimizer))
         _loss_function: torch.nn.MSELoss = torch.nn.MSELoss()
         for e in range(0, self.epoch, 1):
-            print('Epoch: ', e)
             for idx, batch in enumerate(self.iterator):
                 _predictors = torch.autograd.Variable(batch[0])
                 if _predictors.size()[0] != self.batch_size:
@@ -475,7 +501,7 @@ class STC(torch.nn.Module):
                 _prediction = self.auto_encoder(_predictors)
                 _loss = _loss_function(_prediction, _predictors)
                 _loss.backward()
-                self._clip_gradient(clip_value=1e-1)
+                self._clip_gradient(clip_value=1e-1, auto_encoder=True)
                 _optim.step()
 
     @staticmethod
